import ipfshttpclient
import multihash
import hashlib


def get_from_ipfs_and_checkhash(ipfs_client, ipfs_hash_base58, validate=True):
    """
    Get file from IPFS and validate hash
    """
    data = ipfs_client.cat(ipfs_hash_base58)

    if validate:
        block_data = ipfs_client.block.get(ipfs_hash_base58)

        try:
            mh_bytes = multihash.from_b58_string(ipfs_hash_base58)
            decoded = multihash.decode(mh_bytes)

            hash_func_name = decoded.name
            expected_digest = decoded.digest

            if hash_func_name == 'sha2-256': # Standard for IPFS (CIDv0)
                actual_digest = hashlib.sha256(block_data).digest()
            else:
                # Handle other algorithms supported by hashlib if necessary
                h = hashlib.new(hash_func_name.replace('-', ''))
                h.update(block_data)
                actual_digest = h.digest()

            if actual_digest != expected_digest:
                raise Exception("IPFS hash mismatch with data")

        except Exception as e:
            raise ValueError(f"Integrity check failed: {str(e)}") from e

    return data


def get_ipfs_client(config):
    ipfs_endpoint = config.get_ipfs_endpoint()
    return ipfshttpclient.connect(ipfs_endpoint)




"""
Functions for manipulating service metadata

Metadata format:
----------------------------------------------------
version          - used to track format changes (current version is 1)
display_name     - Display name of the service
encoding         - Service encoding (proto or json)
service_type     - Service type (grpc, jsonrpc or process)
service_description - Service description (arbitrary field)
payment_expiration_threshold - Service will reject payments with expiration less
                                than current_block + payment_expiration_threshold.
                               This field should be used by the client with caution.
                               Client should not accept arbitrary payment_expiration_threshold
model_ipfs_hash  - IPFS HASH to the .tar archive of protobuf service specification
mpe_address      - Address of MultiPartyEscrow contract.
                   Client should use it exclusively for cross-checking of mpe_address,
                        (because service can attack via mpe_address)
                   Daemon can use it directly if authenticity of metadata is confirmed
pricing {}      -  Pricing model
         Possible pricing models:
         1. Fixed price
             price_model   - "fixed_price"
             price_in_cogs -  unique fixed price in cogs for all method (1 FET = 10^18 cogs)
             (other pricing models can be easily supported)
groups []       - group is the number of endpoints which shares same payment channel;
                  grouping strategy is defined by service provider;
                  for example service provider can use region name as group name
     group_name - unique name of the group (human readable)
     group_id   - unique id of the group (random 32 byte string in base64 encoding)
     payment_address - Ethereum address to recieve payments
endpoints[]     - address in the off-chain network to provide a service
     group_name
     endpoint   -  unique endpoint identifier (ip:port)

assets {}       -  asset type and its ipfs value/values
"""

import re
import json
import base64

from collections import defaultdict
from enum import Enum

from snet.sdk.utils.utils import is_valid_endpoint


# Supported Asset types
class AssetType(Enum):
    HERO_IMAGE = "hero_image"
    IMAGES = "images"
    DOCUMENTATION = "documentation"
    TERMS_OF_USE = "terms_of_use"

    @staticmethod
    def is_single_value(asset_type):
        if (asset_type == AssetType.HERO_IMAGE.value or asset_type == AssetType.DOCUMENTATION.value
                or asset_type == AssetType.TERMS_OF_USE.value):
            return True


# TODO: we should use some standard solution here
class MPEServiceMetadata:

    def __init__(self):
        """ init with modelIPFSHash """
        self.m = {"version": 1,
                  "display_name": "",
                  "encoding": "grpc",  # grpc by default
                  "service_type": "grpc",  # grpc by default
                  # one week by default (15 sec block,  24*60*60*7/15)
                  "model_ipfs_hash": "",
                  "mpe_address": "",
                  "groups": [],
                  "assets": {},
                  "media": [],
                  "tags": []
                  }

    def set_simple_field(self, f, v):
        if f != "display_name" and f != "encoding" and f != "model_ipfs_hash" and f != "mpe_address" and \
                f != "service_type" and f != "payment_expiration_threshold" and f != "service_description":
            raise Exception("unknown field in MPEServiceMetadata")
        self.m[f] = v

    def set_fixed_price_in_cogs(self, group_name, price):
        if type(price) != int:
            raise Exception("Price should have int type")

        if not self.is_group_name_exists(group_name):
            raise Exception("the group %s is not present" % str(group_name))

        for group in self.m["groups"]:
            if group["group_name"] == group_name:
                is_fixed_price_enabled = False
                # default=True  it will change when we will go live with method level pricing
                if "pricing" in group:
                    for pricing in group['pricing']:
                        if pricing["price_model"] == "fixed_price":
                            is_fixed_price_enabled = True
                            pricing["price_in_cogs"] = price
                    if not is_fixed_price_enabled:
                        group["pricing"].append({"price_model": "fixed_price",
                                                 "price_in_cogs": price, "default": True})
                else:
                    group["pricing"] = [{"price_model": "fixed_price",
                                         "price_in_cogs": price, "default": True}]

    def set_method_price_in_cogs(self, group_name, package_name, service_name, method, price):
        if type(price) != int:
            raise Exception("Price should have int type")

        if not self.is_group_name_exists(group_name):
            raise Exception("the group %s is not present" % str(group_name))

        groups = self.m["groups"]
        for group in groups:
            if group["group_name"] == group_name:

                service_name = service_name
                package_name = package_name
                method_pricing = {"method_name": method,
                                  "price_in_cogs": price}
                pricings = []

                if 'pricings' in group:
                    pricings = group["pricings"]

                fixed_price_method_model_exist = False
                for pricing in pricings:
                    if pricing['price_model'] == 'fixed_price_per_method':
                        fixed_price_method_model_exist = True

                        if 'details' in pricing:
                            fixed_price_method_pricing_for_service_exist = False
                            for detail in pricing['details']:

                                if detail['service_name'] == service_name:
                                    # adding new method pricing for existing service
                                    fixed_price_method_pricing_for_service_exist = True
                                    detail['method_pricing'].append(
                                        method_pricing)

                            if not fixed_price_method_pricing_for_service_exist:
                                # pricing for new method for new service
                                pricing['details'].append({"service_name": service_name,
                                                           "method_pricing": [method_pricing]})
                        else:
                            pricing['details'] = [{"service_name": service_name,
                                                   "method_pricing": [method_pricing]}]

                if not fixed_price_method_model_exist:
                    fixed_price_per_method = {"package_name": package_name,
                                              "price_model": "fixed_price_per_method",
                                              "details": [
                                                  {"service_name": service_name, "method_pricing": [method_pricing]}]}
                    group['pricings'] = [fixed_price_per_method]

    def add_group(self, group_name):
        """ Return new group_id in base64 """
        if self.is_group_name_exists(group_name):
            raise Exception("the group \"%s\" is already present" %
                            str(group_name))

        self.m["groups"] += [{"group_name": group_name}]

    def remove_group(self, group_name):
        for group in self.m["groups"]:
            if group["group_name"] == group_name:
                self.m["groups"].remove(group)

    def get_tags(self):
        tags = []
        if "tags" in self.m:
            tags = self.m["tags"]
        return tags

    def add_tag(self, tag_name):
        if not "tags" in self.m:
            self.m["tags"] = []

        if tag_name in self.m["tags"]:
            print(f"The tag {str(tag_name)} is already present")
            return
        self.m["tags"] += [tag_name]

    def remove_tag(self, tag_name):
        if not "tags" in self.m:
            self.m["tags"] = []

        if tag_name not in self.m["tags"]:
            print(f"The tag {str(tag_name)} is not found")
            return
        self.m["tags"].remove(tag_name)

    def add_asset(self, asset_ipfs_hash, asset_type):
        # Check if we need to validation if same asset type is added twice if we need to add it or replace the existing one

        if 'assets' not in self.m:
            self.m['assets'] = {}

        # hero image will contain the single value
        if AssetType.is_single_value(asset_type):
            self.m['assets'][asset_type] = asset_ipfs_hash

        # images can contain multiple value
        elif asset_type == AssetType.IMAGES.value:
            if asset_type in self.m['assets']:
                self.m['assets'][asset_type].append(asset_ipfs_hash)
            else:
                self.m['assets'][asset_type] = [asset_ipfs_hash]
        else:
            raise Exception("Invalid asset type %s" % asset_type)

    def remove_all_assets(self):
        self.m['assets'] = {}

    def remove_asset(self, asset_type):
        if 'assets' in self.m:
            if AssetType.is_single_value(asset_type):
                self.m['assets'][asset_type] = ""
            elif asset_type == AssetType.IMAGES.value:
                self.m['assets'][asset_type] = []
            else:
                raise Exception("Invalid asset type %s" % asset_type)

    def add_endpoint_to_group(self, group_name, endpoint):
        if re.match("^\w+://", endpoint) is None:
            # TODO: Default to https when our tutorials show setting up a ssl certificate as well
            endpoint = 'http://' + endpoint
        if not is_valid_endpoint(endpoint):
            raise Exception("Endpoint is not a valid URL")
        if not self.is_group_name_exists(group_name):
            raise Exception("the group %s is not present" % str(group_name))
        if endpoint in self.get_all_endpoints_for_group(group_name):
            raise Exception("the endpoint %s is already present" %
                            str(endpoint))

        groups = self.m["groups"]
        for group in groups:
            if group["group_name"] == group_name:
                if 'endpoints' in group:
                    group['endpoints'].append(endpoint)
                else:
                    group['endpoints'] = [endpoint]

    def remove_all_endpoints_for_group(self, group_name):
        if not self.is_group_name_exists(group_name):
            raise Exception("Group name does not exist %s", group_name)

        groups = self.m["groups"]
        for group in groups:
            if group["group_name"] == group_name:
                group["endpoints"] = []

    def is_group_name_exists(self, group_name):
        """ check if group with given name is already exists """
        groups = self.m["groups"]
        for g in groups:
            if g["group_name"] == group_name:
                return True
        return False

    def get_group_by_group_id(self, group_id):
        """ return group with given group_id (return None if it doesn't exist) """
        group_id_base64 = base64.b64encode(group_id).decode('ascii')
        groups = self.m["groups"]
        for g in groups:
            if g["group_id"] == group_id_base64:
                return g
        return None

    def set_free_calls_for_group(self, group_name, free_calls):
        groups = self.m["groups"]
        for g in groups:
            if g["group_name"] == group_name:
                g["free_calls"] = free_calls

    def set_freecall_signer_address(self, group_name, signer_address):
        groups = self.m["groups"]
        for g in groups:
            if g["group_name"] == group_name:
                g["free_call_signer_address"] = signer_address

    def get_json(self):
        return json.dumps(self.m)

    def get_json_pretty(self):
        return json.dumps(self.m, indent=4)

    def set_from_json(self, j):
        # TODO: we probaly should check the  consistensy of loaded json here
        #       check that it contains required fields
        self.m = json.loads(j)
        if not "tags" in self.m:
            self.m["tags"] = []

    def load(self, file_name):
        with open(file_name) as f:
            self.set_from_json(f.read())

    def save_pretty(self, file_name):
        with open(file_name, 'w') as f:
            f.write(self.get_json_pretty())

    def __getitem__(self, key):
        return self.m[key]

    def __contains__(self, key):
        return key in self.m

    def get(self, key, default=None):
        return self.m.get(key, default)

    def get_group_name_nonetrick(self, group_name=None):
        """ In all getter function in case of single payment group, group_name can be None """
        groups = self.m["groups"]
        if len(groups) == 0:
            raise Exception("Cannot find any groups in metadata")
        if not group_name:
            if len(groups) > 1:
                raise Exception(
                    "We have more than one payment group in metadata, so group_name should be specified")
            return groups[0]["group_name"]
        return group_name

    def get_group(self, group_name=None):
        group_name = self.get_group_name_nonetrick(group_name)
        for g in self.m["groups"]:
            if g["group_name"] == group_name:
                return g
        raise Exception('Cannot find group "%s" in metadata' % group_name)

    def get_group_id_base64(self, group_name=None):
        return self.get_group(group_name)["group_id"]

    def get_group_id(self, group_name=None):
        return base64.b64decode(self.get_group_id_base64(group_name))

    def get_payment_address(self, group_name=None):
        return self.get_group(group_name)["payment_address"]

    def add_daemon_address_to_group(self, group_name, daemon_address):
        groups = self.m["groups"]
        if not self.is_group_name_exists(group_name):
            raise Exception('Cannot find group "%s" in metadata' % group_name)
        for group in groups:
            if group["group_name"] == group_name:
                if 'daemon_addresses' in group:
                    group['daemon_addresses'].append(daemon_address)
                else:
                    group['daemon_addresses'] = [daemon_address]

    def remove_all_daemon_addresses_for_group(self, group_name):
        groups = self.m["groups"]
        if not self.is_group_name_exists(group_name):
            raise Exception('Cannot find group "%s" in metadata' % group_name)
        for group in groups:
            if group["group_name"] == group_name:
                group["daemon_addresses"] = []

    def get_all_endpoints_for_group(self, group_name):
        for group in self.m["groups"]:
            if group["group_name"] == group_name:
                if "endpoints" in group:
                    return group["endpoints"]
                return []

    def get_all_group_endpoints(self):
        group_endpoints = {}
        for group in self.m["groups"]:
            if "endpoints" in group:
                group_endpoints[group["group_name"]] = group['endpoints']
        return group_endpoints

    def get_all_endpoints_with_group_name(self):
        endpts_with_grp = defaultdict(list)
        for e in self.m["endpoints"]:
            endpts_with_grp[e['group_name']].append(e['endpoint'])
        return endpts_with_grp

    def get_endpoints_for_group(self, group_name=None):
        group_name = self.get_group_name_nonetrick(group_name)
        return [e["endpoint"] for e in self.m["endpoints"] if e["group_name"] == group_name]

    def add_contributor(self, name, email_id):
        if "contributors" in self.m:
            contributors = self.m["contributors"]
        else:
            contributors = []

        contributors.append(
            {
                "name": name,
                "email_id": email_id
            }
        )
        self.m["contributors"] = contributors

    def remove_contributor_by_email(self, email_id):
        self.m["contributors"] = [
            contributor for contributor in self.m["contributors"] if contributor["email_id"] != email_id
        ]

    def group_init(self, group_name):
        """Required values for creating a new payment group.

        Args:
            group_name: If org contains only 1 payment group -> default_group, ask user for other groups otherwise.

        Raises:
            ValueError: User enters non-integer value for `fixed_price.`
            Exception: User enters same endpoints.
        """
        self.add_group(group_name)
        while True:
            try:
                fixed_price = int(input("Set fixed price: "))
            except ValueError:
                print("Enter a valid integer.")
            else:
                self.set_fixed_price_in_cogs(group_name, fixed_price)
                break
        while True:
            try:
                endpoints = input("Add endpoints as comma separated values: ").split(',')
                if endpoints[0] == "":
                    print("Endpoints required.")
                else:
                    for endpoint in endpoints:
                        self.add_endpoint_to_group(group_name, endpoint.strip())
                    break
            except Exception as e:
                print(e)
        while True:
            daemon_addresses = input("Add daemon addresses as comma separated values: ").split(',')
            if daemon_addresses[0] == "":
                print("Daemon address required.")
            else:
                for daemon_address in daemon_addresses:
                    self.add_daemon_address_to_group(group_name, daemon_address.strip())
                break
        if input('Free calls included? [y/n] ').lower() == 'y':
            self.set_free_calls_for_group(group_name, int(input('free calls: (15) ') or 15))
            self.set_freecall_signer_address(group_name, input('free call signer address: '))

    def add_media(self, url, media_type, hero_img=False):
        """Add new individual media to service metadata."""
        if 'media' not in self.m:
            self.m['media'] = []
        individual_media = {}
        if hero_img:
            assert (media_type == 'image'), f"{media_type.upper()} media-type cannot be a hero-image."
            assert (not self._is_asset_type_exists()), "Hero-image already exists (only 1 unique hero-image allowed.)"
            individual_media['asset_type'] = AssetType.HERO_IMAGE.value  # Dependency with AssetType, fix if obsolete
        if len(self.m['media']) == 0:
            individual_media['order'] = 1
        else:
            individual_media['order'] = self.m['media'][-1]['order'] + 1
        individual_media['url'] = url
        individual_media['file_type'] = media_type
        if media_type == 'image':
            individual_media['alt_text'] = 'hover_on_the_image_text'
        else:
            individual_media['alt_text'] = 'hover_on_the_video_url'
        self.m['media'].append(individual_media)

    def remove_media(self, order):
        """Remove individual media from service metadata using unique order key."""
        assert (len(self.m['media']) > 0), "No media content to remove."
        assert (order > 0), "Order of individual media starts from 1."
        del_position = -1
        for i in range(len(self.m['media'])):
            if order == self.m['media'][i]['order']:
                del self.m['media'][i]
                del_position = i
                break
        if del_position == -1:
            raise Exception(f"Media with order: {order} not found.")
        for i in range(del_position, len(self.m['media'])):
            self.m['media'][i]['order'] -= 1

    def remove_all_media(self):
        """Remove all individual media from metadata."""
        self.m['media'].clear()

    def swap_media_order(self, move_from, move_to):
        """Swap orders of two different media given their individual orders (move_from, move_to)."""
        assert (len(self.m['media']) + 1 > move_from > 0), f"Order {move_from} out of bounds."
        assert (len(self.m['media']) + 1 > move_to > 0), f"Order {move_to} out of bounds."
        self.m['media'][move_to - 1], self.m['media'][move_from - 1] = self.m['media'][move_from - 1], self.m['media'][
            move_to - 1]
        self.m['media'][move_to - 1]['order'], self.m['media'][move_from - 1]['order'] = self.m['media'][move_from - 1][
            'order'], \
            self.m['media'][move_to - 1][
                'order']

    def change_media_order(self):
        """Mini REPL to change order of all individual media"""
        order_range = range(1, len(self.m['media']) + 1)
        available_orders = list(order_range)
        for individual_media in self.m['media']:
            print(f"File Type: {individual_media['file_type']}, Current Order: {individual_media['order']}")
            while True:
                try:
                    new_order = int(input(f"Enter new order for {individual_media['url']}: "))
                except ValueError:
                    print("Error: Order entered not a number. Try again.")
                else:
                    if new_order in available_orders:
                        individual_media['order'] = new_order
                        available_orders.remove(new_order)
                        break
                    elif new_order not in order_range:
                        print(
                            f"Media array contains only {len(self.m['media'])} items. Enter order between [{order_range.start}, {order_range.stop - 1}]")
                    else:
                        print(f"Order already taken. Available orders: {available_orders}")
        self.m['media'].sort(key=lambda x: x['order'])

    def _is_asset_type_exists(self):
        """Return boolean on whether asset type already exists"""
        media = self.m['media']
        for individual_media in media:
            if 'asset_type' in individual_media:
                return True
        return False

    def add_description(self):
        if 'service_description' not in self.m:
            self.m['service_description'] = {
                "url": input("user guide url: "),
                "long_description": input("service long description: "),
                "short_description": input("service short description: ")
            }


def load_mpe_service_metadata(f) -> MPEServiceMetadata:
    metadata = MPEServiceMetadata()
    metadata.load(f)
    return metadata


def mpe_service_metadata_from_json(j) -> MPEServiceMetadata:
    metadata = MPEServiceMetadata()
    metadata.set_from_json(j)
    return metadata



from enum import Enum
from typing import Any


class ModelMethodMessage(Enum):
    CreateModel = "create_model"
    ValidateModelPrice = "validate_model_price"
    TrainModelPrice = "train_model_price"
    DeleteModel = "delete_model"
    GetTrainingMetadata = "get_training_metadata"
    GetAllModels = "get_all_models"
    GetModel = "get_model"
    UpdateModel = "update_model"
    GetMethodMetadata = "get_method_metadata"
    UploadAndValidate = "upload_and_validate"
    ValidateModel = "validate_model"
    TrainModel = "train_model"


class ModelStatus(Enum):
    CREATED = 0
    VALIDATING = 1
    VALIDATED = 2
    TRAINING = 3
    READY_TO_USE = 4
    ERRORED = 5
    DELETED = 6


def to_string(obj: Any):
    res_str = ""
    for key, value in obj.__dict__.items():
        key = key.split("__")[1].replace("_", " ")
        res_str += f"{key}: {value}\n"
    return res_str


class Model:
    def __init__(self, model_response):
        self.__model_id = model_response.model_id
        self.__status = ModelStatus(model_response.status)
        self.__created_date = model_response.created_date
        self.__updated_date = model_response.updated_date
        self.__name = model_response.name
        self.__description = model_response.description
        self.__grpc_method_name = model_response.grpc_method_name
        self.__grpc_service_name = model_response.grpc_service_name
        self.__address_list = model_response.address_list
        self.__is_public = model_response.is_public
        self.__training_data_link = model_response.training_data_link
        self.__created_by_address = model_response.created_by_address
        self.__updated_by_address = model_response.updated_by_address

    def __str__(self):
        return to_string(self)

    @property
    def model_id(self):
        return self.__model_id

    @property
    def status(self):
        return self.__status

    @property
    def created_date(self):
        return self.__created_date

    @property
    def updated_date(self):
        return self.__updated_date

    @property
    def name(self):
        return self.__name

    @property
    def description(self):
        return self.__description

    @property
    def grpc_method_name(self):
        return self.__grpc_method_name

    @property
    def grpc_service_name(self):
        return self.__grpc_service_name

    @property
    def address_list(self):
        return self.__address_list

    @property
    def is_public(self):
        return self.__is_public

    @property
    def training_data_link(self):
        return self.__training_data_link

    @property
    def created_by_address(self):
        return self.__created_by_address

    @property
    def updated_by_address(self):
        return self.__updated_by_address


class TrainingMetadata:
    def __init__(self,
                 training_enabled: bool,
                 training_in_proto: bool,
                 training_methods: Any):

        self.__training_enabled = training_enabled
        self.__training_in_proto = training_in_proto
        self.__training_methods = {}

        services_methods = dict(training_methods)
        for k, v in services_methods.items():
            self.__training_methods[k] = [value.string_value for value in v.values]

    def __str__(self):
        return to_string(self)

    @property
    def training_enabled(self):
        return self.__training_enabled

    @property
    def training_in_proto(self):
        return self.__training_in_proto

    @property
    def training_methods(self):
        return self.__training_methods


class MethodMetadata:
    def __init__(self,
                 default_model_id: str,
                 max_models_per_user: int,
                 dataset_max_size_mb: int,
                 dataset_max_count_files: int,
                 dataset_max_size_single_file_mb: int,
                 dataset_files_type: str,
                 dataset_type: str,
                 dataset_description: str):

        self.__default_model_id = default_model_id
        self.__max_models_per_user = max_models_per_user
        self.__dataset_max_size_mb = dataset_max_size_mb
        self.__dataset_max_count_files = dataset_max_count_files
        self.__dataset_max_size_single_file_mb = dataset_max_size_single_file_mb
        self.__dataset_files_type = dataset_files_type
        self.__dataset_type = dataset_type
        self.__dataset_description = dataset_description

    def __str__(self):
        return to_string(self)

    @property
    def default_model_id(self):
        return self.__default_model_id

    @property
    def max_models_per_user(self):
        return self.__max_models_per_user

    @property
    def dataset_max_size_mb(self):
        return self.__dataset_max_size_mb

    @property
    def dataset_max_count_files(self):
        return self.__dataset_max_count_files

    @property
    def dataset_max_size_single_file_mb(self):
        return self.__dataset_max_size_single_file_mb

    @property
    def dataset_files_type(self):
        return self.__dataset_files_type

    @property
    def dataset_type(self):
        return self.__dataset_type

    @property
    def dataset_description(self):
        return self.__dataset_description





import web3
import importlib
from eth_account.messages import defunct_hash_message

from snet.sdk.utils.utils import RESOURCES_PATH, add_to_path

class PaymentChannel:
    def __init__(self, channel_id, w3, account, payment_channel_state_service_client, mpe_contract):
        self.channel_id = channel_id
        self.web3 = w3
        self.account = account
        self.mpe_contract = mpe_contract
        self.payment_channel_state_service_client = payment_channel_state_service_client
        self.state = {
            "nonce": 0,
            "last_signed_amount": 0
        }

    def add_funds(self, amount):
        return self.mpe_contract.channel_add_funds(self.account, self.channel_id, amount)

    def extend_expiration(self, expiration):
        return self.mpe_contract.channel_extend(self.account, self.channel_id, expiration)

    def extend_and_add_funds(self, expiration, amount):
        return self.mpe_contract.channel_extend_and_add_funds(self.account, self.channel_id, expiration, amount)

    def sync_state(self):
        channel_blockchain_data = self.mpe_contract.contract.functions.channels(self.channel_id).call()
        (current_nonce, last_signed_amount) = self._get_current_channel_state()
        nonce = channel_blockchain_data[0]
        total_amount = channel_blockchain_data[5]
        expiration = channel_blockchain_data[6]
        available_amount = total_amount - last_signed_amount
        self.state = {
            "current_nonce": current_nonce,
            "last_signed_amount": last_signed_amount,
            "nonce": nonce,
            "total_amount": total_amount,
            "expiration": expiration,
            "available_amount": available_amount
        }

    def _get_current_channel_state(self):
        stub = self.payment_channel_state_service_client
        current_block_number = self.web3.eth.get_block("latest").number
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256", "uint256"],
            [
                "__get_channel_state",
                web3.Web3.to_checksum_address(self.mpe_contract.contract.address),
                self.channel_id,current_block_number
            ]
        )
        signature = self.web3.eth.account._sign_hash(defunct_hash_message(message), self.account.signer_private_key).signature
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2 = importlib.import_module("state_service_pb2")
        request = state_service_pb2.ChannelStateRequest(
            channel_id=web3.Web3.to_bytes(self.channel_id),
            signature=bytes(signature),
            current_block=current_block_number
        )
        response = stub.GetChannelState(request)
        return int.from_bytes(response.current_nonce, byteorder="big"),int.from_bytes(response.current_signed_amount, byteorder="big")



import os
from pathlib import Path

from snet.sdk.storage_provider.storage_provider import StorageProvider
from snet.sdk.utils.utils import compile_proto


class ClientLibGenerator:
    def __init__(self, metadata_provider: StorageProvider, org_id: str,
                 service_id: str, protodir: Path | None = None):
        self._metadata_provider: StorageProvider = metadata_provider
        self.org_id: str = org_id
        self.service_id: str = service_id
        self.language: str = "python"
        self.protodir: Path = (protodir if protodir else
                               Path.home().joinpath(".snet"))
        self.generate_directories_by_params()

    def generate_client_library(self) -> None:
        try:
            self.receive_proto_files()
            compilation_result = compile_proto(entry_path=self.protodir,
                                               codegen_dir=self.protodir,
                                               target_language=self.language,
                                               add_training=self.training_added())
            if compilation_result:
                print(f'client libraries for service with id "{self.service_id}" '
                      f'in org with id "{self.org_id}" '
                      f'generated at {self.protodir}')
        except Exception as e:
            print(str(e))

    def generate_directories_by_params(self) -> None:
        if not self.protodir.is_absolute():
            self.protodir = Path.cwd().joinpath(self.protodir)
        self.create_service_client_libraries_path()

    def create_service_client_libraries_path(self) -> None:
        self.protodir = self.protodir.joinpath(self.org_id,
                                               self.service_id,
                                               self.language)
        self.protodir.mkdir(parents=True, exist_ok=True)

    def receive_proto_files(self) -> None:
        metadata = self._metadata_provider.fetch_service_metadata(
            org_id=self.org_id,
            service_id=self.service_id
        )
        service_api_source = (metadata.get("service_api_source") or
                              metadata.get("model_ipfs_hash"))

        # Receive proto files
        if self.protodir.exists():
            self._metadata_provider.fetch_and_extract_proto(
                service_api_source,
                self.protodir
            )
        else:
            raise Exception("Directory for storing proto files is not found")

    def training_added(self) -> bool:
        files = os.listdir(self.protodir)
        for file in files:
            if ".proto" not in file:
                continue
            with open(self.protodir.joinpath(file), "r") as f:
                proto_text = f.read()
            if 'import "training.proto";' in proto_text:
                return True
        return False



import importlib
import os
import sys
import warnings
from enum import Enum

import google.protobuf.internal.api_implementation

from google.protobuf import symbol_database as _symbol_database

from snet.sdk.storage_provider.service_metadata import MPEServiceMetadata

with warnings.catch_warnings():
    # Suppress the eth-typing package`s warnings related to some new networks
    warnings.filterwarnings(
        "ignore",
        "Network .* does not have a valid ChainId. eth-typing should be "
        "updated with the latest networks.",
        UserWarning
    )

    import web3

from snet.contracts import get_contract_object
from snet.sdk.account import Account
from snet.sdk.config import Config
from snet.sdk.client_lib_generator import ClientLibGenerator
from snet.sdk.mpe.mpe_contract import MPEContract
from snet.sdk.mpe.payment_channel_provider import PaymentChannelProvider
from snet.sdk.payment_strategies.default_payment_strategy import *
from snet.sdk.service_client import ServiceClient
from snet.sdk.storage_provider.storage_provider import StorageProvider
from snet.sdk.custom_typing import ModuleName, ServiceStub
from snet.sdk.utils.utils import (
    bytes32_to_str,
    find_file_by_keyword,
    type_converter
)

google.protobuf.internal.api_implementation.Type = lambda: 'python'
_sym_db = _symbol_database.Default()
_sym_db.RegisterMessage = lambda x: None


class PaymentStrategyType(Enum):
    PAID_CALL = PaidCallPaymentStrategy
    FREE_CALL = FreeCallPaymentStrategy
    PREPAID_CALL = PrePaidPaymentStrategy
    DEFAULT = DefaultPaymentStrategy


class SnetSDK:
    """Base Snet SDK"""

    def __init__(self, sdk_config: Config, metadata_provider=None):
        self._sdk_config = sdk_config
        self._metadata_provider = metadata_provider

        # Instantiate Ethereum client
        eth_rpc_endpoint = self._sdk_config["eth_rpc_endpoint"]
        eth_rpc_request_kwargs = self._sdk_config.get("eth_rpc_request_kwargs")

        provider = web3.HTTPProvider(endpoint_uri=eth_rpc_endpoint,
                                     request_kwargs=eth_rpc_request_kwargs)

        self.web3 = web3.Web3(provider)

        # Get MPE contract address from config if specified;
        # mostly for local testing
        _mpe_contract_address = self._sdk_config.get("mpe_contract_address",
                                                     None)
        if _mpe_contract_address is None:
            self.mpe_contract = MPEContract(self.web3)
        else:
            self.mpe_contract = MPEContract(self.web3, _mpe_contract_address)

        # Get Registry contract address from config if specified;
        # mostly for local testing
        _registry_contract_address = self._sdk_config.get(
            "registry_contract_address",
            None
        )
        if _registry_contract_address is None:
            self.registry_contract = get_contract_object(self.web3, "Registry")
        else:
            self.registry_contract = get_contract_object(
                self.web3,
                "Registry",
                _registry_contract_address
            )

        if self._metadata_provider is None:
            self._metadata_provider = StorageProvider(self._sdk_config,
                                                      self.registry_contract)

        self.account = Account(self.web3, sdk_config, self.mpe_contract)
        self.payment_channel_provider = PaymentChannelProvider(
            self.web3,
            self.mpe_contract
        )

    def create_service_client(self,
                              org_id: str,
                              service_id: str,
                              group_name: str=None,
                              payment_strategy: PaymentStrategy = None,
                              payment_strategy_type: PaymentStrategyType=PaymentStrategyType.DEFAULT,
                              options=None,
                              concurrent_calls: int = 1):

        # Create and instance of the Config object,
        # so we can create an instance of ClientLibGenerator
        self.lib_generator = ClientLibGenerator(self._metadata_provider,
                                                org_id, service_id)

        # Download the proto file and generate stubs if needed
        force_update = self._sdk_config.get('force_update', False)
        if force_update:
            self.lib_generator.generate_client_library()
        else:
            path_to_pb_files = self.lib_generator.protodir
            pb_2_file_name = find_file_by_keyword(path_to_pb_files,
                                                  keyword="pb2.py",
                                                  exclude=["training"])
            pb_2_grpc_file_name = find_file_by_keyword(path_to_pb_files,
                                                       keyword="pb2_grpc.py",
                                                       exclude=["training"])
            if not pb_2_file_name or not pb_2_grpc_file_name:
                print("Generating client library...")
                self.lib_generator.generate_client_library()

        if options is None:
            options = dict()
        options['concurrency'] = self._sdk_config.get("concurrency", True)
        options['concurrent_calls'] = concurrent_calls

        if payment_strategy is None:
            payment_strategy = payment_strategy_type.value()

        service_metadata = self._metadata_provider.enhance_service_metadata(
            org_id, service_id
        )
        group = self._get_service_group_details(service_metadata, group_name)

        service_stubs = self.get_service_stub()

        pb2_module = self.get_module_by_keyword(keyword="pb2.py")
        _service_client = ServiceClient(org_id, service_id, service_metadata,
                                        group, service_stubs,
                                        payment_strategy,
                                        options, self.mpe_contract,
                                        self.account, self.web3, pb2_module,
                                        self.payment_channel_provider,
                                        self.lib_generator.protodir,
                                        self.lib_generator.training_added())
        return _service_client

    def get_service_stub(self) -> list[ServiceStub]:
        path_to_pb_files = str(self.lib_generator.protodir)
        module_name = self.get_module_by_keyword(keyword="pb2_grpc.py")
        sys.path.append(path_to_pb_files)
        try:
            grpc_file = importlib.import_module(module_name)
            properties_and_methods_of_grpc_file = dir(grpc_file)
            service_stubs = []
            for elem in properties_and_methods_of_grpc_file:
                if 'Stub' in elem:
                    service_stubs.append(getattr(grpc_file, elem))
            return [ServiceStub(service_stub) for service_stub in service_stubs]

        except Exception as e:
            raise Exception(f"Error importing module: {e}")

    def get_module_by_keyword(self, keyword: str) -> ModuleName:
        path_to_pb_files = self.lib_generator.protodir
        file_name = find_file_by_keyword(path_to_pb_files,
                                         keyword,
                                         exclude=["training"])
        module_name = os.path.splitext(file_name)[0]
        return ModuleName(module_name)

    def get_service_metadata(self, org_id, service_id):
        return self._metadata_provider.fetch_service_metadata(org_id,
                                                              service_id)

    def _get_first_group(self, service_metadata: MPEServiceMetadata) -> dict:
        return service_metadata['groups'][0]

    def _get_group_by_group_name(self,
                                 service_metadata: MPEServiceMetadata,
                                 group_name: str) -> dict:
        for group in service_metadata['groups']:
            if group['group_name'] == group_name:
                return group
        return {}

    def _get_service_group_details(self,
                                   service_metadata: MPEServiceMetadata,
                                   group_name: str) -> dict:
        if len(service_metadata['groups']) == 0:
            raise Exception("No Groups found for given service, "
                            "Please add group to the service")

        if group_name is None:
            return self._get_first_group(service_metadata)

        return self._get_group_by_group_name(service_metadata, group_name)

    def get_organization_list(self) -> list:
        org_list = self.registry_contract.functions.listOrganizations().call()
        organization_list = []
        for idx, org_id in enumerate(org_list):
            organization_list.append(bytes32_to_str(org_id))
        return organization_list

    def get_services_list(self, org_id: str) -> list:
        found, org_service_list = (
            self.registry_contract.functions
            .listServicesForOrganization(type_converter("bytes32")(org_id))
            .call())
        if not found:
            raise Exception(f"Organization with id={org_id} doesn't exist!")
        org_service_list = list(map(bytes32_to_str, org_service_list))
        return org_service_list





class Config:
    def __init__(self,
                 private_key,
                 eth_rpc_endpoint,
                 wallet_index=0,
                 ipfs_endpoint=None,
                 concurrency=True,
                 force_update=False,
                 mpe_contract_address=None,
                 token_contract_address=None,
                 registry_contract_address=None,
                 signer_private_key=None):
        self.__config = {
            "private_key": private_key,
            "eth_rpc_endpoint": eth_rpc_endpoint,
            "wallet_index": wallet_index,
            "ipfs_endpoint": (ipfs_endpoint if ipfs_endpoint
                              else "/dns/ipfs.singularitynet.io/tcp/80/"),
            "concurrency": concurrency,
            "force_update": force_update,
            "mpe_contract_address": mpe_contract_address,
            "token_contract_address": token_contract_address,
            "registry_contract_address": registry_contract_address,
            "signer_private_key": signer_private_key,
            "lighthouse_token": " "
        }

    def __getitem__(self, key):
        return self.__config[key]

    def get(self, key, default=None):
        return self.__config.get(key, default)

    def get_ipfs_endpoint(self):
        return self["ipfs_endpoint"]



import importlib

import grpc
import web3

from snet.sdk.payment_strategies.payment_strategy import PaymentStrategy
from snet.sdk.utils.utils import RESOURCES_PATH, add_to_path

class FreeCallPaymentStrategy(PaymentStrategy):

    def __init__(self):
        self._user_address = None
        self._free_call_token = None
        self._token_expiration_block = None

    def get_free_calls_available(self, service_client) -> int:
        if not self._user_address:
            self._user_address = service_client.account.signer_address

        current_block_number = service_client.get_current_block_number()

        if (not self._free_call_token or
                not self._token_expiration_block or
                current_block_number > self._token_expiration_block):
            self._free_call_token, self._token_expiration_block = self.get_free_call_token_details(service_client)

        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2 = importlib.import_module("state_service_pb2")

        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2_grpc = importlib.import_module("state_service_pb2_grpc")

        signature, _ = self.generate_signature(service_client, current_block_number)
        request = state_service_pb2.FreeCallStateRequest(
            address=self._user_address,
            free_call_token=self._free_call_token,
            signature=signature,
            current_block=current_block_number
        )

        channel = service_client.get_grpc_base_channel()
        stub = state_service_pb2_grpc.FreeCallStateServiceStub(channel)

        try:
            response = stub.GetFreeCallsAvailable(request)
            return response.free_calls_available
        except grpc.RpcError as e:
            if self._user_address:
                print(f"Warning: {e.details()}")
            return 0

    def get_payment_metadata(self, service_client) -> list:
        if self.get_free_calls_available(service_client) <= 0:
            raise Exception(f"Free calls limit for address {self._user_address} has expired. Please use another payment strategy")
        signature, current_block_number = self.generate_signature(service_client)
        metadata = [("snet-free-call-auth-token-bin", self._free_call_token),
                    ("snet-payment-type", "free-call"),
                    ("snet-free-call-user-address", self._user_address),
                    ("snet-current-block-number", str(current_block_number)),
                    ("snet-payment-channel-signature-bin", signature)]

        return metadata

    def generate_signature(self, service_client, current_block_number=None, with_token=True) -> tuple[bytes, int]:
        if not current_block_number:
            current_block_number = service_client.get_current_block_number()
        org_id, service_id, group_id, _ = service_client.get_service_details()

        message_types = ["string", "string", "string", "string", "string", "uint256", "bytes32"]
        message_values = ["__prefix_free_trial", self._user_address, org_id, service_id, group_id,
                          current_block_number, self._free_call_token]

        if not with_token:
            message_types = message_types[:-1]
            message_values = message_values[:-1]

        message = web3.Web3.solidity_keccak(message_types, message_values)
        return service_client.generate_signature(message), current_block_number

    def get_free_call_token_details(self, service_client, current_block_number=None) -> tuple[bytes, int]:

        signature, current_block_number = self.generate_signature(service_client, current_block_number, with_token=False)

        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2 = importlib.import_module("state_service_pb2")

        request = state_service_pb2.GetFreeCallTokenRequest(
            address=self._user_address,
            signature=signature,
            current_block=current_block_number
        )

        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2_grpc = importlib.import_module("state_service_pb2_grpc")

        channel = service_client.get_grpc_base_channel()
        stub = state_service_pb2_grpc.FreeCallStateServiceStub(channel)
        response = stub.GetFreeCallToken(request)

        return response.token, response.token_expiration_block
        


import web3

from snet.sdk.payment_strategies.paidcall_payment_strategy import PaidCallPaymentStrategy


class TrainingPaymentStrategy(PaidCallPaymentStrategy):
    def __init__(self):
        super().__init__()
        self._call_price = -1
        self._train_model_id = ""

    def get_price(self, service_client=None) -> int:
        if self._call_price == -1:
            raise Exception("Training call price not set")
        return self._call_price

    def set_price(self, call_price: int) -> None:
        self._call_price = call_price

    def get_model_id(self):
        return self._train_model_id

    def set_model_id(self, model_id: str):
        self._train_model_id = model_id

    def get_payment_metadata(self, service_client) -> list[tuple[str, str]]:
        channel = self.select_channel(service_client)
        amount = channel.state["last_signed_amount"] + int(self.get_price(service_client))
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256", "uint256", "uint256"],
            ["__MPE_claim_message", service_client.mpe_address, channel.channel_id,
             channel.state["nonce"],
             amount]
        )
        signature = service_client.generate_signature(message)

        metadata = [
            ("snet-payment-type", "train-call"),
            ("snet-payment-channel-id", str(channel.channel_id)),
            ("snet-payment-channel-nonce", str(channel.state["nonce"])),
            ("snet-payment-channel-amount", str(amount)),
            ("snet-train-model-id", self.get_model_id()),
            ("snet-payment-channel-signature-bin", signature)
        ]

        return metadata



import base64
import importlib
import re
import os
from pathlib import Path
from typing import Any

from eth_typing import BlockNumber
import grpc
from hexbytes import HexBytes
import web3
from eth_account.messages import defunct_hash_message
from urllib.parse import urlparse

from snet.sdk import generic_client_interceptor, FreeCallPaymentStrategy
from snet.sdk.account import Account
from snet.sdk.mpe.mpe_contract import MPEContract
from snet.sdk.mpe.payment_channel import PaymentChannel
from snet.sdk.mpe.payment_channel_provider import PaymentChannelProvider
from snet.sdk.payment_strategies.prepaid_payment_strategy import PrePaidPaymentStrategy
from snet.sdk.resources.root_certificate import certificate
from snet.sdk.storage_provider.service_metadata import MPEServiceMetadata
from snet.sdk.custom_typing import ModuleName, ServiceStub
from snet.sdk.utils.utils import (RESOURCES_PATH, add_to_path,
                                  find_file_by_keyword)
from snet.sdk.training.training import Training
from snet.sdk.training.exceptions import NoTrainingException
from snet.sdk.utils.call_utils import create_intercept_call_func


class ServiceClient:
    def __init__(
        self,
        org_id: str,
        service_id: str,
        service_metadata: MPEServiceMetadata,
        group: dict,
        service_stubs: list[ServiceStub],
        payment_strategy,
        options: dict,
        mpe_contract: MPEContract,
        account: Account,
        sdk_web3: web3.Web3,
        pb2_module: ModuleName,
        payment_channel_provider: PaymentChannelProvider,
        path_to_pb_files: Path,
        training_added: bool = False
    ):
        self.org_id = org_id
        self.service_id = service_id
        self.service_metadata = service_metadata
        self.group = group
        self.payment_strategy = payment_strategy
        if isinstance(payment_strategy, PrePaidPaymentStrategy):
            self.payment_strategy.set_concurrent_calls(options["concurrent_calls"])
        self.options = options
        self.mpe_address = mpe_contract.contract.address
        self.account = account
        self.sdk_web3 = sdk_web3
        self.pb2_module = (importlib.import_module(pb2_module)
                                if isinstance(pb2_module, str)
                                else pb2_module)
        self.payment_channel_provider = payment_channel_provider
        self.path_to_pb_files = path_to_pb_files

        self.expiry_threshold: int = self.group["payment"]["payment_expiration_threshold"]
        self.__base_grpc_channel = self._get_grpc_channel()
        _intercept_call_func = create_intercept_call_func(self.payment_strategy.get_payment_metadata, self)
        self.grpc_channel = grpc.intercept_channel(
            self.__base_grpc_channel,
            generic_client_interceptor.create(_intercept_call_func)
        )
        self.service_stubs = service_stubs
        self.payment_channel_state_service_client = self._generate_payment_channel_state_service_client()
        self.payment_channels = []
        self.last_read_block: int = 0
        self.__training = Training(self, training_added)

    def call_rpc(self, rpc_name: str, message_class: str, **kwargs) -> Any:
        service = self._get_service_stub(rpc_name)
        if "model_id" in kwargs:
            kwargs["model_id"] = self._get_training_model_id(kwargs["model_id"])
        rpc_method = getattr(service, rpc_name)
        request = getattr(self.pb2_module, message_class)(**kwargs)
        return rpc_method(request)

    def _get_payment_expiration_threshold_for_group(self):
        pass

    def _get_service_stub(self, rpc_name: str) -> Any:
        for service_stub in self.service_stubs:
            grpc_stub = self._generate_grpc_stub(service_stub)
            if hasattr(grpc_stub, rpc_name):
                return grpc_stub
        raise Exception(f"Service stub for {rpc_name} not found")

    def _generate_grpc_stub(self, service_stub: ServiceStub) -> Any:
        grpc_channel = self.__base_grpc_channel
        disable_blockchain_operations: bool = self.options.get(
            "disable_blockchain_operations",
            False
        )
        if not disable_blockchain_operations:
            grpc_channel = self.grpc_channel
        stub_instance = service_stub(grpc_channel)
        return stub_instance

    def get_grpc_base_channel(self) -> grpc.Channel:
        return self.__base_grpc_channel

    def _get_grpc_channel(self) -> grpc.Channel:
        endpoint = self.options.get("endpoint", None)
        if endpoint is None:
            endpoint = self.service_metadata.get_all_endpoints_for_group(self.group["group_name"])[0]
        endpoint_object = urlparse(endpoint)
        if endpoint_object.port is not None:
            channel_endpoint = endpoint_object.hostname + ":" + str(endpoint_object.port)
        else:
            channel_endpoint = endpoint_object.hostname

        if endpoint_object.scheme == "http":
            return grpc.insecure_channel(channel_endpoint)
        elif endpoint_object.scheme == "https":
            return grpc.secure_channel(channel_endpoint,
                                       grpc.ssl_channel_credentials(root_certificates=certificate))
        else:
            raise ValueError('Unsupported scheme in service metadata ("{}")'.format(endpoint_object.scheme))

    def _filter_existing_channels_from_new_payment_channels(
        self,
        new_payment_channels: list[PaymentChannel]
    ) -> list[PaymentChannel]:
        new_channels_to_be_added = []

        # need to change this logic ,use maps to manage channels so that we can easily navigate it
        for new_payment_channel in new_payment_channels:
            existing_channel = False
            for existing_payment_channel in self.payment_channels:
                if new_payment_channel.channel_id == existing_payment_channel.channel_id:
                    existing_channel = True
                    break

            if not existing_channel:
                new_channels_to_be_added.append(new_payment_channel)

        return new_channels_to_be_added

    def load_open_channels(self) -> list[PaymentChannel]:
        current_block_number = self.sdk_web3.eth.block_number
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        new_payment_channels = (
            self.payment_channel_provider.get_past_open_channels(
                self.account, payment_address,
                group_id, self.payment_channel_state_service_client
            )
        )
        filter_new_channels = self._filter_existing_channels_from_new_payment_channels(new_payment_channels)
        self.payment_channels = (self.payment_channels + filter_new_channels)
        self.last_read_block = current_block_number
        return self.payment_channels

    def get_current_block_number(self) -> BlockNumber:
        return self.sdk_web3.eth.block_number

    def update_channel_states(self) -> list[PaymentChannel]:
        for channel in self.payment_channels:
            channel.sync_state()
        return self.payment_channels

    def default_channel_expiration(self) -> int:
        current_block_number = self.sdk_web3.eth.get_block("latest").number
        return current_block_number + self.expiry_threshold

    def _generate_payment_channel_state_service_client(self) -> Any:
        grpc_channel = self.__base_grpc_channel
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service = importlib.import_module("state_service_pb2_grpc")
        return state_service.PaymentChannelStateServiceStub(grpc_channel)

    def open_channel(self, amount: int, expiration: int) -> PaymentChannel:
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        return self.payment_channel_provider.open_channel(
            self.account, amount, expiration, payment_address,
            group_id, self.payment_channel_state_service_client
        )

    def deposit_and_open_channel(self, amount: int,
                                 expiration: int) -> PaymentChannel:
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        return self.payment_channel_provider.deposit_and_open_channel(
            self.account, amount, expiration, payment_address,
            group_id, self.payment_channel_state_service_client
        )

    def get_price(self) -> int:
        return self.group["pricing"][0]["price_in_cogs"]

    def generate_signature(self, message: bytes) -> bytes:
        return bytes(self.sdk_web3.eth.account._sign_hash(
            defunct_hash_message(message), self.account.signer_private_key
        ).signature)

    def generate_training_signature(self, text: str, address: str,
                                    block_number: BlockNumber) -> HexBytes:
        address = web3.Web3.to_checksum_address(address)
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256"],
            [text, address, block_number]
        )
        return self.sdk_web3.eth.account._sign_hash(
            defunct_hash_message(message), self.account.signer_private_key
        ).signature

    def get_service_details(self) -> tuple[str, str, str, str]:
        return (self.org_id,
                self.service_id,
                self.group["group_id"],
                self.service_metadata.get_all_endpoints_for_group(
                    self.group["group_name"]
                )[0])

    @property
    def training(self) -> Training:
        if not self.__training.is_enabled:
            raise NoTrainingException(self.org_id, self.service_id)
        return self.__training

    def _get_training_model_id(self, model_id: str) -> Any:
        return self.training.get_model_id_object(model_id)

    def get_concurrency_flag(self) -> bool:
        return self.options.get('concurrency', True)

    def get_concurrency_token_and_channel(self) -> tuple[str, PaymentChannel]:
        return self.payment_strategy.get_concurrency_token_and_channel(self)

    def get_concurrent_calls(self):
        return self.options.get('concurrent_calls', 1)

    def set_concurrency_token_and_channel(self, token: str,
                                          channel: PaymentChannel) -> None:
        self.payment_strategy.concurrency_token = token
        self.payment_strategy.channel = channel

    def get_services_and_messages_info(self) -> tuple[dict, dict]:
        # Get proto file filepath and open it
        proto_file_name = find_file_by_keyword(directory=self.path_to_pb_files,
                                               keyword=".proto", exclude=["training"])
        proto_filepath = os.path.join(self.path_to_pb_files, proto_file_name)
        with open(proto_filepath, 'r') as file:
            proto_content = file.read()
        # Regular expression patterns to match services, methods,
        # messages and fields
        service_pattern = re.compile(r'service\s+(\w+)\s*{')
        rpc_pattern = re.compile(
            r'rpc\s+(\w+)\s*\((\w+)\)\s+returns\s+\((\w+)\)'
        )
        message_pattern = re.compile(r'message\s+(\w+)\s*{')
        field_pattern = re.compile(r'\s*(\w+)\s+(\w+)\s*=\s*\d+\s*;')

        services = {}
        messages = {}
        current_service = None
        current_message = None

        for line in proto_content.splitlines():
            # Match a service definition
            service_match = service_pattern.search(line)
            if service_match:
                current_service = service_match.group(1)
                services[current_service] = []
                continue

            # Match an RPC method inside a service
            if current_service:
                rpc_match = rpc_pattern.search(line)
                if rpc_match:
                    method_name = rpc_match.group(1)
                    input_type = rpc_match.group(2)
                    output_type = rpc_match.group(3)
                    services[current_service].append((method_name, input_type,
                                                      output_type))

            # Match a message definition
            message_match = message_pattern.search(line)
            if message_match:
                current_message = message_match.group(1)
                messages[current_message] = []
                continue

            # Match a field inside a message
            if current_message:
                field_match = field_pattern.search(line)
                if field_match:
                    field_type = field_match.group(1)
                    field_name = field_match.group(2)
                    messages[current_message].append((field_type, field_name))

        return services, messages

    def get_services_and_messages_info_as_pretty_string(self) -> str:
        services, messages = self.get_services_and_messages_info()

        string_output = ""
        # Prettify the parsed services and their methods
        for service, methods in services.items():
            string_output += f"Service: {service}\n"
            for method_name, input_type, output_type in methods:
                string_output += (f"  Method: {method_name},"
                                  f" Input: {input_type},"
                                  f" Output: {output_type}\n")

        # Prettify the messages and their fields
        for message, fields in messages.items():
            string_output += f"Message: {message}\n"
            for field_type, field_name in fields:
                string_output += f"  Field: {field_type} {field_name}\n"
        return string_output

    def get_free_calls_available(self) -> int:
        payment_strategy = self.payment_strategy
        if not isinstance(payment_strategy, FreeCallPaymentStrategy):
            payment_strategy = FreeCallPaymentStrategy()

        return payment_strategy.get_free_calls_available(self)



from snet.sdk.payment_strategies.freecall_payment_strategy import FreeCallPaymentStrategy
from snet.sdk.payment_strategies.paidcall_payment_strategy import PaidCallPaymentStrategy
from snet.sdk.payment_strategies.prepaid_payment_strategy import PrePaidPaymentStrategy
from snet.sdk.payment_strategies.payment_strategy import PaymentStrategy


class DefaultPaymentStrategy(PaymentStrategy):

    def __init__(self):
        self.channel = None

    def set_channel(self, channel):
        self.channel = channel

    def get_payment_metadata(self, service_client):
        free_call_payment_strategy = FreeCallPaymentStrategy()

        if free_call_payment_strategy.get_free_calls_available(service_client) > 0:
            metadata = free_call_payment_strategy.get_payment_metadata(service_client)
        else:
            if service_client.get_concurrency_flag():
                concurrent_calls = service_client.get_concurrent_calls()
                payment_strategy = PrePaidPaymentStrategy(concurrent_calls)
                metadata = payment_strategy.get_payment_metadata(service_client)
            else:
                payment_strategy = PaidCallPaymentStrategy()
                metadata = payment_strategy.get_payment_metadata(service_client)

        return metadata

    def get_price(self, service_client):
        pass

    def get_concurrency_token_and_channel(self, service_client):
        concurrent_calls = service_client.get_concurrent_calls()
        payment_strategy = PrePaidPaymentStrategy(concurrent_calls)
        return payment_strategy.get_concurrency_token_and_channel(service_client)



from snet.contracts import get_contract_deployment_block, get_contract_object


class MPEContract:
    def __init__(self, w3, address=None):
        self.web3 = w3
        if address is None:
            self.contract = get_contract_object(self.web3, "MultiPartyEscrow")
        else:
            self.contract = get_contract_object(self.web3, "MultiPartyEscrow", address)

    def balance(self, address):
        return self.contract.functions.balances(address).call()

    def deposit(self, account, amount_in_cogs):
        return account.send_transaction(self.contract.functions.deposit, amount_in_cogs)

    def open_channel(self, account, payment_address, group_id, amount, expiration):
        return account.send_transaction(self.contract.functions.openChannel, account.signer_address, payment_address,
                                        group_id, amount, expiration)

    def deposit_and_open_channel(self, account, payment_address, group_id, amount, expiration):
        already_approved_amount = account.allowance()
        if amount > already_approved_amount:
            account.approve_transfer(amount)
        return account.send_transaction(self.contract.functions.depositAndOpenChannel, account.signer_address,
                                        payment_address, group_id, amount, expiration)

    def channel_add_funds(self, account, channel_id, amount):
        self._fund_escrow_account(account, amount)
        return account.send_transaction(self.contract.functions.channelAddFunds, channel_id, amount)

    def channel_extend(self, account, channel_id, expiration):
        return account.send_transaction(self.contract.functions.channelExtend, channel_id, expiration)

    def channel_extend_and_add_funds(self, account, channel_id, expiration, amount):
        self._fund_escrow_account(account, amount)
        return account.send_transaction(self.contract.functions.channelExtendAndAddFunds, channel_id, expiration,
                                        amount)

    def _fund_escrow_account(self, account, amount):
        current_escrow_balance = self.balance(account.address)
        if amount > current_escrow_balance:
            account.deposit_to_escrow_account(amount - current_escrow_balance)



from grpc import RpcError


class WrongDatasetException(Exception):
    def __init__(self, errors: list[str]):
        self.errors = errors
        exception_msg = "Dataset check failed:\n"
        for check in errors:
            exception_msg += f"\t{check}\n"
        super().__init__(exception_msg)


class WrongMethodException(Exception):
    def __init__(self, method_name: str):
        super().__init__(f"Method with name {method_name} not found!")


class NoTrainingException(Exception):
    def __init__(self, org_id: str, service_id: str):
        super().__init__(f"Training is not implemented for the service with org_id={org_id} and service_id={service_id}!")


class GRPCException(RpcError):
    def __init__(self, error: RpcError):
        super().__init__(f"An error occurred during the grpc call: {error}.")


class NoSuchModelException(Exception):
    def __init__(self, model_id: str):
        super().__init__(f"Model with id {model_id} not found!")




import importlib

import grpc
import web3

from snet.sdk.utils.utils import RESOURCES_PATH, add_to_path


class ConcurrencyManager:
    def __init__(self, concurrent_calls: int=1):
        self.__concurrent_calls: int = concurrent_calls
        self.__token: str = ''
        self.__planned_amount: int = 0
        self.__used_amount: int = 0

    @property
    def concurrent_calls(self) -> int:
        return self.__concurrent_calls

    @concurrent_calls.setter
    def concurrent_calls(self, concurrent_calls: int):
        self.__concurrent_calls = concurrent_calls

    def get_token(self, service_client, channel, service_call_price):
        if len(self.__token) == 0:
            self.__token = self.__get_token(service_client, channel, service_call_price)
        elif self.__used_amount >= self.__planned_amount:
            self.__token = self.__get_token(service_client, channel, service_call_price, new_token=True)
        return self.__token

    def __get_token(self, service_client, channel, service_call_price, new_token=False):
        if not new_token:
            amount = channel.state["last_signed_amount"]
            if amount != 0:
                try:
                    token_reply = self.__get_token_for_amount(service_client, channel, amount)
                    planned_amount = token_reply.planned_amount
                    used_amount = token_reply.used_amount
                    if planned_amount - used_amount > 0:
                        self.__used_amount = used_amount
                        self.__planned_amount = planned_amount
                        return token_reply.token
                except grpc.RpcError as e:
                    if e.details() != "Unable to retrieve planned Amount ":
                        raise

        amount = channel.state["last_signed_amount"] + service_call_price
        token_reply = self.__get_token_for_amount(service_client, channel, amount)
        self.__used_amount = token_reply.used_amount
        self.__planned_amount = token_reply.planned_amount
        return token_reply.token

    def __get_stub_for_get_token(self, service_client):
        grpc_channel = service_client.get_grpc_base_channel()
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            token_service_pb2_grpc = importlib.import_module("token_service_pb2_grpc")
        return token_service_pb2_grpc.TokenServiceStub(grpc_channel)

    def __get_token_for_amount(self, service_client, channel, amount):
        nonce = channel.state["nonce"]
        stub = self.__get_stub_for_get_token(service_client)
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            token_service_pb2 = importlib.import_module("token_service_pb2")
        current_block_number = service_client.sdk_web3.eth.get_block("latest").number
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256", "uint256", "uint256"],
            ["__MPE_claim_message", service_client.mpe_address, channel.channel_id, nonce, amount]
        )
        mpe_signature = service_client.generate_signature(message)
        message = web3.Web3.solidity_keccak(
            ["bytes", "uint256"],
            [mpe_signature, current_block_number]
        )
        sign_mpe_signature = service_client.generate_signature(message)

        request = token_service_pb2.TokenRequest(
            channel_id=channel.channel_id, current_nonce=nonce, signed_amount=amount,
            signature=bytes(sign_mpe_signature), claim_signature=bytes(mpe_signature),
            current_block=current_block_number)
        token_reply = stub.GetToken(request)
        return token_reply

    def record_successful_call(self):
        self.__used_amount += 1



import json
import sys
import importlib.resources
from urllib.parse import urlparse
from pathlib import Path, PurePath
import os
import tarfile
import io

import web3
from eth_typing import BlockNumber
from grpc_tools.protoc import main as protoc

from snet import sdk

RESOURCES_PATH = PurePath(os.path.dirname(sdk.__file__)).joinpath("resources")


def safe_address_converter(a):
    if not web3.Web3.is_checksum_address(a):
        raise Exception("%s is not is not a valid Ethereum checksum address" % a)
    return a


def type_converter(t):
    if t.endswith("[]"):
        return lambda x: list(map(type_converter(t.replace("[]", "")), json.loads(x)))
    else:
        if "int" in t:
            return lambda x: web3.Web3.to_int(text=x)
        elif "bytes32" in t:
            return lambda x: web3.Web3.to_bytes(text=x).ljust(32, b"\0") if not x.startswith(
                "0x") else web3.Web3.to_bytes(hexstr=x).ljust(32, b"\0")
        elif "byte" in t:
            return lambda x: web3.Web3.to_bytes(text=x) if not x.startswith("0x") else web3.Web3.to_bytes(hexstr=x)
        elif "address" in t:
            return safe_address_converter
        else:
            return str


def bytes32_to_str(b):
    return b.rstrip(b"\0").decode("utf-8")


def compile_proto(
    entry_path: Path,
    codegen_dir: Path,
    proto_file: str | None = None,
    target_language: str = "python",
    add_training: bool = False
) -> bool:
    try:
        if not os.path.exists(codegen_dir):
            os.makedirs(codegen_dir)
        proto_include = importlib.resources.files('grpc_tools') / '_proto'

        compiler_args = [
            "-I{}".format(entry_path),
            "-I{}".format(proto_include)
        ]

        if add_training:
            training_include = RESOURCES_PATH.joinpath("proto", "training")
            compiler_args.append("-I{}".format(training_include))

        if target_language == "python":
            compiler_args.insert(0, "protoc")
            compiler_args.append("--python_out={}".format(codegen_dir))
            compiler_args.append("--grpc_python_out={}".format(codegen_dir))
            compiler = protoc
        else:
            raise Exception("We only support python target language for proto compiling")

        if proto_file:
            compiler_args.append(str(proto_file))
        else:
            compiler_args.extend([str(p) for p in entry_path.glob("**/*.proto")])

        if add_training:
            compiler_args.append(str(training_include.joinpath("training.proto")))

        if not compiler(compiler_args):
            return True
        else:
            return False

    except Exception as e:
        print(e)
        return False


def is_valid_endpoint(url):
    """
    Just ensures the url has a scheme (http/https), and a net location (IP or domain name).
    Can make more advanced or do on-network tests if needed, but this is really just to catch obvious errors.
    >>> is_valid_endpoint("https://34.216.72.29:6206")
    True
    >>> is_valid_endpoint("blahblah")
    False
    >>> is_valid_endpoint("blah://34.216.72.29")
    False
    >>> is_valid_endpoint("http://34.216.72.29:%%%")
    False
    >>> is_valid_endpoint("http://192.168.0.2:9999")
    True
    """
    try:
        result = urlparse(url)
        if result.port:
            _port = int(result.port)
        return (
                all([result.scheme, result.netloc]) and
                result.scheme in ['http', 'https']
        )
    except ValueError:
        return False


def normalize_private_key(private_key):
    if private_key.startswith("0x"):
        private_key = bytes(bytearray.fromhex(private_key[2:]))
    else:
        private_key = bytes(bytearray.fromhex(private_key))
    return private_key


def get_address_from_private(private_key):
    return web3.Account.from_key(private_key).address


def get_current_block_number() -> BlockNumber:
    return web3.Web3().eth.block_number


class add_to_path:
    def __init__(self, path):
        self.path = path

    def __enter__(self):
        sys.path.insert(0, self.path)

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            sys.path.remove(self.path)
        except ValueError:
            pass


def find_file_by_keyword(directory, keyword, exclude=None):
    if exclude is None:
        exclude = []
    for root, dirs, files in os.walk(directory):
        for file in files:
            if keyword in file and all(e not in file for e in exclude):
                return file


def bytesuri_to_hash(s, to_decode=True):
    if to_decode:
        s = s.rstrip(b"\0").decode('ascii')
    if s.startswith("ipfs://"):
        return "ipfs", s[7:]
    elif s.startswith("filecoin://"):
        return "filecoin", s[11:]
    else:
        raise Exception("We support only ipfs and filecoin uri in Registry")


def safe_extract_proto(spec_tar, protodir):
    """
    Tar files might be dangerous (see https://bugs.python.org/issue21109,
    and https://docs.python.org/3/library/tarfile.html, TarFile.extractall warning)
    we extract only simple files
    """
    with tarfile.open(fileobj=io.BytesIO(spec_tar)) as f:
        for m in f.getmembers():
            if os.path.dirname(m.name) != "":
                raise Exception(
                    "tarball has directories. We do not support it.")
            if not m.isfile():
                raise Exception(
                    "tarball contains %s which is not a file" % m.name)
            fullname = os.path.join(protodir, m.name)
            if os.path.exists(fullname):
                os.remove(fullname)
                print(f"{fullname} removed.")
        # now it is safe to call extractall
        f.extractall(path=protodir)



import web3
from lighthouseweb3 import Lighthouse
import json

from snet.sdk.utils.ipfs_utils import get_ipfs_client, get_from_ipfs_and_checkhash
from snet.sdk.utils.utils import bytesuri_to_hash, safe_extract_proto
from snet.sdk.storage_provider.service_metadata import MPEServiceMetadata, mpe_service_metadata_from_json

class StorageProvider(object):
    def __init__(self, config, registry_contract):
        self._registry_contract = registry_contract
        self._ipfs_client = get_ipfs_client(config)
        self.lighthouse_client = Lighthouse(config["lighthouse_token"])

    def fetch_org_metadata(self,org_id):
        org = web3.Web3.to_bytes(text=org_id).ljust(32, b"\0")

        found, _, org_metadata_uri, _, _, _ = self._registry_contract.functions.getOrganizationById(org).call()
        if found is not True:
            raise Exception('Organization with org ID "{}" not found '.format(org_id))

        org_provider_type, org_metadata_hash = bytesuri_to_hash(org_metadata_uri)

        if org_provider_type == "ipfs":
            org_metadata_json = get_from_ipfs_and_checkhash(self._ipfs_client, org_metadata_hash)
        else:
            org_metadata_json, _ = self.lighthouse_client.download(org_metadata_hash)
        org_metadata = json.loads(org_metadata_json)

        return org_metadata

    def fetch_service_metadata(self, org_id: str,
                               service_id: str) -> MPEServiceMetadata:
        org = web3.Web3.to_bytes(text=org_id).ljust(32, b"\0")
        service = web3.Web3.to_bytes(text=service_id).ljust(32, b"\0")

        found, _, service_metadata_uri = (
            self._registry_contract.functions.getServiceRegistrationById(
                org,
                service
            ).call()
        )
        if found is not True:
            raise Exception(f"No service '{service_id}' "
                            f"found in organization '{org_id}'")

        service_provider_type, service_metadata_hash = bytesuri_to_hash(
            s=service_metadata_uri
        )

        if service_provider_type == "ipfs":
            service_metadata_json = get_from_ipfs_and_checkhash(
                self._ipfs_client,
                service_metadata_hash
            )
        else:
            service_metadata_json, _ = self.lighthouse_client.download(
                cid=service_metadata_hash
            )
        service_metadata = mpe_service_metadata_from_json(service_metadata_json)

        return service_metadata

    def enhance_service_metadata(self,org_id,service_id):
        service_metadata = self.fetch_service_metadata(org_id, service_id)
        org_metadata = self.fetch_org_metadata(org_id)

        org_group_map = {}
        for group in org_metadata['groups']:
            org_group_map[group['group_name']] = group

        for group in service_metadata.m['groups']:
            # merge service group with org_group
            group['payment'] = org_group_map[group['group_name']]['payment']

        return service_metadata

    def fetch_and_extract_proto(self, service_api_source, protodir):
        try:
            proto_provider_type, service_api_source = bytesuri_to_hash(service_api_source, to_decode=False)
        except Exception:
            proto_provider_type = "ipfs"

        if proto_provider_type == "ipfs":
            spec_tar = get_from_ipfs_and_checkhash(self._ipfs_client, service_api_source)
        else:
            spec_tar, _ = self.lighthouse_client.download(service_api_source)

        safe_extract_proto(spec_tar, protodir)




import collections
import grpc


class _ClientCallDetails(
    collections.namedtuple(
        '_ClientCallDetails',
        ('method', 'timeout', 'metadata', 'credentials')),
    grpc.ClientCallDetails):
    pass


def create_intercept_call_func(get_metadata_func: callable, service_client) -> callable:
    def intercept_call(client_call_details, request_iterator, request_streaming, response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.extend(get_metadata_func(service_client))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return intercept_call



from pathlib import Path
from setuptools import setup
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install
from setuptools.command.build_py import build_py as _build_py
from grpc_tools import protoc
import importlib.resources


def install_and_compile_proto():
    """
    Compiles protobuf files directly.
    """
    proto_dir = Path(__file__).absolute().parent.joinpath(
        "snet", "sdk", "resources", "proto")

    grpc_protos_include = str(importlib.resources.files('grpc_tools').joinpath('_proto'))

    print(f"Proto directory: {proto_dir}")
    print(f"Grpc include directory: {grpc_protos_include}")

    if not proto_dir.exists():
        print(f"Warning: Proto directory not found at {proto_dir}")
        return

    # glob('*.proto') is non-recursive. It will NOT look inside subfolders.
    for fn in proto_dir.glob('*.proto'):
        print(f"Compiling protobuf: {fn}")

        command = [
            'grpc_tools.protoc',
            f'-I{proto_dir}',
            f'-I{grpc_protos_include}',
            f'--python_out={proto_dir}',
            f'--grpc_python_out={proto_dir}',
            str(fn)
        ]

        if protoc.main(command) != 0:
            print(f"Error: Failed to compile {fn}")
            raise RuntimeError(f"Protocol buffer compilation failed for {fn}")


class build_py(_build_py):
    """
    This is the hook used by 'python -m build'.
    """
    def run(self):
        self.execute(install_and_compile_proto, (), msg="Compile protocol buffers")
        _build_py.run(self)


class develop(_develop):
    """Post-installation for development mode (pip install -e .)."""
    def run(self):
        self.execute(install_and_compile_proto, (), msg="Compile protocol buffers")
        _develop.run(self)


class install(_install):
    """Post-installation for legacy installation mode."""
    def run(self):
        self.execute(install_and_compile_proto, (), msg="Compile protocol buffers")
        _install.run(self)


setup(
    cmdclass={
        'develop': develop,
        'install': install,
        'build_py': build_py,
    },
)


import json

import web3

from snet.contracts import get_contract_object
from snet.sdk.config import Config
from snet.sdk.mpe.mpe_contract import MPEContract
from snet.sdk.utils.utils import (get_address_from_private,
                                  normalize_private_key)

DEFAULT_GAS = 300000
TRANSACTION_TIMEOUT = 500


class TransactionError(Exception):
    """
    Raised when an Ethereum transaction receipt has a status of 0.
    Can provide a custom message. Optionally includes receipt
    """

    def __init__(self, message, receipt=None):
        super().__init__(message)
        self.message = message
        self.receipt = receipt

    def __str__(self):
        return self.message


class Account:
    def __init__(self, w3: web3.Web3, config: Config,
                 mpe_contract: MPEContract):
        self.config: Config = config
        self.web3: web3.Web3 = w3
        self.mpe_contract: MPEContract = mpe_contract
        _token_contract_address: str | None = self.config.get(
            "token_contract_address", None
        )
        if _token_contract_address is None:
            self.token_contract = get_contract_object(
                self.web3, "FetchToken")
        else:
            self.token_contract = get_contract_object(
                self.web3, "FetchToken", _token_contract_address)

        if config.get("private_key") is not None:
            self.private_key = normalize_private_key(config.get("private_key"))
        if config.get("signer_private_key") is not None:
            self.signer_private_key = normalize_private_key(
                config.get("signer_private_key")
            )
        else:
            self.signer_private_key = self.private_key
        self.address = get_address_from_private(self.private_key)
        self.signer_address = get_address_from_private(self.signer_private_key)
        self.nonce = 0

    def _get_nonce(self):
        nonce = self.web3.eth.get_transaction_count(self.address)
        if self.nonce >= nonce:
            nonce = self.nonce + 1
        self.nonce = nonce
        return nonce

    def _get_gas_price(self):
        gas_price = self.web3.eth.gas_price
        if gas_price <= 15000000000:
            gas_price += gas_price * 1 / 3
        elif gas_price > 15000000000 and gas_price <= 50000000000:
            gas_price += gas_price * 1 / 5
        elif gas_price > 50000000000 and gas_price <= 150000000000:
            gas_price += 7000000000
        elif gas_price > 150000000000:
            gas_price += gas_price * 1 / 10
        return int(gas_price)

    def _send_signed_transaction(self, contract_fn, *args):
        transaction = contract_fn(*args).build_transaction({
            "chainId": int(self.web3.net.version),
            "gas": DEFAULT_GAS,
            "gasPrice": self._get_gas_price(),
            "nonce": self._get_nonce()
        })
        signed_txn = self.web3.eth.account.sign_transaction(
            transaction, private_key=self.private_key)
        return self.web3.to_hex(
            self.web3.eth.send_raw_transaction(signed_txn.raw_transaction)
        )

    def send_transaction(self, contract_fn, *args):
        txn_hash = self._send_signed_transaction(contract_fn, *args)
        return self.web3.eth.wait_for_transaction_receipt(txn_hash,
                                                          TRANSACTION_TIMEOUT)

    def _parse_receipt(self, receipt, event, encoder=json.JSONEncoder):
        if receipt.status == 0:
            raise TransactionError("Transaction failed", receipt)
        else:
            return json.dumps(dict(event().processReceipt(receipt)[0]["args"]),
                              cls=encoder)

    def escrow_balance(self):
        return self.mpe_contract.balance(self.address)

    def deposit_to_escrow_account(self, amount_in_cogs):
        already_approved = self.allowance()
        if amount_in_cogs > already_approved:
            self.approve_transfer(amount_in_cogs)
        return self.mpe_contract.deposit(self, amount_in_cogs)

    def approve_transfer(self, amount_in_cogs):
        return self.send_transaction(self.token_contract.functions.approve,
                                     self.mpe_contract.contract.address,
                                     amount_in_cogs)

    def allowance(self):
        return self.token_contract.functions.allowance(
            self.address,
            self.mpe_contract.contract.address
        ).call()



from snet.sdk.concurrency_manager import ConcurrencyManager
from snet.sdk.payment_strategies.payment_strategy import PaymentStrategy


class PrePaidPaymentStrategy(PaymentStrategy):

    def __init__(self, concurrent_calls: int=1, block_offset: int = 240, call_allowance: int = 1):
        self.concurrency_manager = ConcurrencyManager(concurrent_calls)
        self.block_offset = block_offset
        self.call_allowance = call_allowance

    def get_price(self, service_client):
        return service_client.get_price() * self.concurrency_manager.concurrent_calls

    def set_concurrent_calls(self, concurrent_calls):
        self.concurrency_manager.concurrent_calls = concurrent_calls

    def get_payment_metadata(self, service_client):
        channel = self.select_channel(service_client)
        token = self.concurrency_manager.get_token(service_client, channel, self.get_price(service_client))
        metadata = [
            ("snet-payment-type", "prepaid-call"),
            ("snet-payment-channel-id", str(channel.channel_id)),
            ("snet-payment-channel-nonce", str(channel.state["nonce"])),
            ("snet-prepaid-auth-token-bin", bytes(token, 'UTF-8'))
        ]
        return metadata

    def get_concurrency_token_and_channel(self, service_client):
        channel = self.select_channel(service_client)
        token = self.concurrency_manager.get_token(service_client, channel, self.get_price(service_client))
        return token, channel

    def select_channel(self, service_client):
        account = service_client.account
        service_client.load_open_channels()
        service_client.update_channel_states()
        payment_channels = service_client.payment_channels
        service_call_price = self.get_price(service_client)
        extend_channel_fund = service_call_price * self.call_allowance
        mpe_balance = account.escrow_balance()
        default_expiration = service_client.default_channel_expiration()

        if len(payment_channels) < 1:
            if service_call_price > mpe_balance:
                payment_channel = service_client.deposit_and_open_channel(service_call_price,
                                                                          default_expiration + self.block_offset)
            else:
                payment_channel = service_client.open_channel(service_call_price,
                                                              default_expiration + self.block_offset)
        else:
            payment_channel = payment_channels[0]

        if self.__has_sufficient_funds(payment_channel, service_call_price) \
                and not self.__is_valid(payment_channel, default_expiration):
            payment_channel.extend_expiration(default_expiration + self.block_offset)

        elif not self.__has_sufficient_funds(payment_channel, service_call_price) and \
                self.__is_valid(payment_channel, default_expiration):
            payment_channel.add_funds(extend_channel_fund)

        elif not self.__has_sufficient_funds(payment_channel, service_call_price) and \
                not self.__is_valid(payment_channel, default_expiration):
            payment_channel.extend_and_add_funds(default_expiration + self.block_offset, extend_channel_fund)

        return payment_channel

    @staticmethod
    def __has_sufficient_funds(channel, amount):
        return channel.state["available_amount"] >= amount

    @staticmethod
    def __is_valid(channel, expiry):
        return channel.state["expiration"] >= expiry



from pathlib import PurePath, Path
import os
from zipfile import ZipFile
from typing import Any
import grpc
import importlib

from snet.sdk import generic_client_interceptor
from snet.sdk.payment_strategies.training_payment_strategy import TrainingPaymentStrategy
from snet.sdk.utils.call_utils import create_intercept_call_func
from snet.sdk.utils.utils import add_to_path, RESOURCES_PATH
from snet.sdk.training.exceptions import (
    WrongDatasetException,
    WrongMethodException,
    GRPCException,
    NoSuchModelException
)
from snet.sdk.training.responses import (
    ModelStatus,
    Model,
    TrainingMetadata,
    MethodMetadata,
    ModelMethodMessage
)


class Training:
    def __init__(self, service_client, training_added=False):
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            self.training_daemon = importlib.import_module("training_daemon_pb2")
            self.training_daemon_grpc = importlib.import_module("training_daemon_pb2_grpc")
            self.training = importlib.import_module("training_pb2")
            
        self.service_client = service_client
        self.is_enabled = training_added and self._check_training()
        self.payment_strategy = TrainingPaymentStrategy()

    def get_model_id_object(self, model_id: str) -> Any:
        return self.training.ModelID(model_id=model_id)

    """FREE METHODS TO CALL"""

    def create_model(self, method_name: str,
                     model_name: str,
                     model_description: str="",
                     is_public_accessible: bool=False,
                     addresses_with_access: list[str]=None) -> Model:
        if addresses_with_access is None:
            addresses_with_access = []

        service_name, method_name = self._check_method_name(method_name)
        new_model = self.training.NewModel(name=model_name,
                                      description=model_description,
                                      grpc_method_name=method_name,
                                      grpc_service_name=service_name,
                                      is_public=is_public_accessible,
                                      address_list=addresses_with_access,
                                      organization_id="",
                                      service_id="",
                                      group_id="")
        auth_details = self._get_auth_details(ModelMethodMessage.CreateModel)
        create_model_request = self.training_daemon.NewModelRequest(authorization=auth_details,
                                                               model=new_model)
        response = self._call_method("create_model",
                                     request_data=create_model_request)
        model = Model(response)

        return model

    def validate_model_price(self, model_id: str) -> int:
        auth_details = self._get_auth_details(ModelMethodMessage.ValidateModelPrice)
        validate_model_price_request = self.training_daemon.AuthValidateRequest(
            authorization=auth_details,
            model_id=model_id,
            training_data_link=""
        )

        try:
            response = self._call_method("validate_model_price",
                                         request_data=validate_model_price_request)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e

        return response.price

    def train_model_price(self, model_id: str) -> int:
        auth_details = self._get_auth_details(ModelMethodMessage.TrainModelPrice)
        common_request = self.training_daemon.CommonRequest(authorization=auth_details,
                                                       model_id=model_id)
        try:
            response = self._call_method("train_model_price",
                                         request_data=common_request)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e

        return response.price

    def delete_model(self, model_id: str) -> ModelStatus:
        auth_details = self._get_auth_details(ModelMethodMessage.DeleteModel)
        common_request = self.training_daemon.CommonRequest(authorization=auth_details,
                                                       model_id=model_id)
        try:
            response = self._call_method("delete_model",
                                         request_data=common_request)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e

        return ModelStatus(response.status)

    def get_training_metadata(self) -> TrainingMetadata:
        empty_request = self.training_daemon.google_dot_protobuf_dot_empty__pb2.Empty()
        response = self._call_method("get_training_metadata",
                                     request_data=empty_request)

        training_metadata = TrainingMetadata(training_enabled=response.trainingEnabled,
                                             training_in_proto=response.trainingInProto,
                                             training_methods=response.trainingMethods)

        return training_metadata

    def get_all_models(self, statuses: list[ModelStatus]=None,
                       is_public: bool=None,
                       grpc_method_name: str="",
                       grpc_service_name: str="",
                       model_name: str="",
                       created_by_address: str="") -> list[Model]:
        if statuses is not None:
            statuses = [getattr(self.training.Status, status.value) for status in statuses]
        auth_details = self._get_auth_details(ModelMethodMessage.GetAllModels)
        all_models_request = self.training_daemon.AllModelsRequest(
            authorization=auth_details,
            statuses=statuses,
            is_public=is_public,
            grpc_method_name=grpc_method_name,
            grpc_service_name=grpc_service_name,
            name=model_name,
            created_by_address=created_by_address,
            page_size=0,
            page=0
        )
        response = self._call_method("get_all_models",
                                     request_data=all_models_request)
        models = []
        for model in response.list_of_models:
            models.append(Model(model))

        return models

    def get_model(self, model_id: str) -> Model:
        auth_details = self._get_auth_details(ModelMethodMessage.GetModel)
        common_request = self.training_daemon.CommonRequest(authorization=auth_details,
                                                       model_id=model_id)
        try:
            response = self._call_method("get_model",
                                         request_data=common_request)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e
        model = Model(response)

        return model

    def get_method_metadata(self, method_name: str, model_id: str= "") -> MethodMetadata:

        if model_id:
            requirements_request = self.training_daemon.MethodMetadataRequest(
                grpc_method_name="",
                grpc_service_name="",
                model_id=model_id
            )
        else:
            service_name, method_name = self._check_method_name(method_name)
            requirements_request = self.training_daemon.MethodMetadataRequest(
                grpc_method_name=method_name,
                grpc_service_name=service_name,
                model_id=""
            )
        response = self._call_method("get_method_metadata",
                                     request_data=requirements_request)

        method_metadata = MethodMetadata(default_model_id = response.default_model_id,
                                         max_models_per_user = response.max_models_per_user,
                                         dataset_max_size_mb = response.dataset_max_size_mb,
                                         dataset_max_count_files = response.dataset_max_count_files,
                                         dataset_max_size_single_file_mb = response.dataset_max_size_single_file_mb,
                                         dataset_files_type = response.dataset_files_type,
                                         dataset_type = response.dataset_type,
                                         dataset_description = response.dataset_description)
        return method_metadata

    def update_model(self, model_id: str,
                     model_name: str=None,
                     description: str=None,
                     addresses_with_access: list[str]=None) -> Model:
        auth_details = self._get_auth_details(ModelMethodMessage.UpdateModel)
        update_model_request = self.training_daemon.UpdateModelRequest(
            authorization=auth_details,model_id=model_id,
            model_name=model_name,
            description=description,
            address_list=addresses_with_access
        )

        try:
            response = self._call_method("update_model",
                                         request_data=update_model_request)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e

        model = Model(response)

        return model

    """PAID METHODS TO CALL"""

    def upload_and_validate(self, model_id: str,
                            zip_path: str | Path | PurePath, price: int) -> ModelStatus:
        f = open(zip_path, 'rb')
        file_size = os.path.getsize(zip_path)

        file_name = os.path.basename(zip_path)
        file_size = file_size
        batch_size = 1024*1024
        batch_count = file_size // batch_size
        if file_size % batch_size != 0:
            batch_count += 1

        self._check_dataset(model_id, zip_path)

        auth_details = self._get_auth_details(ModelMethodMessage.UploadAndValidate)

        def request_iter(file):
            batch = file.read(batch_size)
            batch_number = 1
            while batch:
                upload_input = self.training.UploadInput(
                    model_id = model_id,
                    data = batch,
                    file_name = file_name,
                    file_size = file_size,
                    batch_size = batch_size,
                    batch_number = batch_number,
                    batch_count = batch_count
                )
                yield self.training_daemon.UploadAndValidateRequest(
                    authorization=auth_details,
                    upload_input=upload_input
                )
                batch = file.read(batch_size)
                batch_number += 1

        self.payment_strategy.set_price(price)
        self.payment_strategy.set_model_id(model_id)

        try:
            response = self._call_method("upload_and_validate",
                                         request_data=request_iter(f),
                                         paid=True)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e
        finally:
            f.close()

        return ModelStatus(response.status)

    def train_model(self, model_id: str, price: int) -> ModelStatus:
        auth_details = self._get_auth_details(ModelMethodMessage.TrainModel)
        common_request = self.training_daemon.CommonRequest(authorization=auth_details,
                                                       model_id=model_id)
        self.payment_strategy.set_price(price)
        self.payment_strategy.set_model_id(model_id)

        try:
            response = self._call_method("train_model",
                                         request_data=common_request,
                                         paid=True)
        except GRPCException as e:
            if "unable to access model" in str(e):
                raise NoSuchModelException(model_id)
            else:
                raise e

        return ModelStatus(response.status)

    """PRIVATE METHODS"""

    def _call_method(self, method_name: str,
                     request_data,
                     paid=False) -> Any:
        try:
            stub = self._get_training_stub(paid=paid)
            response = getattr(stub, method_name)(request_data)
            return response
        except grpc.RpcError as e:
            raise GRPCException(e)

    def _get_training_stub(self, paid=False) -> Any:
        grpc_channel = self.service_client.get_grpc_base_channel()
        if paid:
            grpc_channel = self._get_grpc_channel(grpc_channel)
        return self.training_daemon_grpc.DaemonStub(grpc_channel)

    def _get_auth_details(self, method_msg: ModelMethodMessage) -> Any:
        current_block_number = self.service_client.get_current_block_number()
        address = self.service_client.account.address
        signature = self.service_client.generate_training_signature(method_msg.value,
                                                                    address,
                                                                    current_block_number)
        auth_details = self.training_daemon.AuthorizationDetails(
            signature=bytes(signature),
            current_block=current_block_number,
            signer_address=address,
            message=method_msg.value
        )
        return auth_details

    def _check_method_name(self, method_name: str) -> tuple[str, str]:
        services_methods, _ = self.service_client.get_services_and_messages_info()
        for service, methods in services_methods.items():
            for method in methods:
                if method[0] == method_name:
                    return service, method[0]
        raise WrongMethodException(method_name)

    def _check_training(self) -> bool:
        try:
            service_methods = self.get_training_metadata().training_methods
        except GRPCException as e:
            return False
        if len(service_methods.keys()) == 0:
            return False
        n_methods = 0
        for service, methods in service_methods.items():
            n_methods += len(methods)
        if n_methods == 0:
            return False
        else:
            return True

    def _check_dataset(self, model_id: str, zip_path: str | Path | PurePath) -> None:
        method_metadata = self.get_method_metadata("", model_id)
        max_size_mb = method_metadata.dataset_max_size_mb
        max_count_files = method_metadata.dataset_max_count_files
        max_size_mb_single = method_metadata.dataset_max_size_single_file_mb
        file_types = method_metadata.dataset_files_type
        file_types = [i for i in file_types.replace(' ', '').split(',')]
        if file_types[0] == '':
            file_types = []

        failed_checks = []
        zip_file = ZipFile(zip_path)

        if os.path.getsize(zip_path) > max_size_mb * 1024 * 1024 > 0:
            failed_checks.append(f"Too big dataset size: "
                                 f"{os.path.getsize(zip_path)//1024/1024} MB > {max_size_mb} MB")

        files_list = zip_file.infolist()
        if len(files_list) > max_count_files > 0:
            failed_checks.append(f"Too many files: {len(files_list)} > {max_count_files}")

        for file_info in files_list:
            _, extension = os.path.splitext(file_info.filename)
            extension = extension[1:] if extension.startswith('.') else extension
            if len(file_types) > 0 and extension not in file_types:
                failed_checks.append(f"Wrong file type: `{extension}` in file: "
                                     f"`{file_info.filename}`. Allowed file types: "
                                     f"{', '.join(file_types)}")
            if file_info.file_size > max_size_mb_single * 1024 * 1024 > 0:
                failed_checks.append(f"Too big file `{file_info.filename}` size: "
                                     f"{file_info.file_size // 1024 / 1024} MB > "
                                     f"{max_size_mb_single} MB")

        if len(failed_checks) > 0:
            raise WrongDatasetException(failed_checks)

    def _get_grpc_channel(self, base_channel: grpc.Channel) -> grpc.Channel:
        intercept_call_func = create_intercept_call_func(self.payment_strategy.get_payment_metadata,
                                                         self.service_client)
        grpc_channel = grpc.intercept_channel(
            base_channel,
            generic_client_interceptor.create(intercept_call_func)
        )
        return grpc_channel



from pathlib import Path

from web3._utils.events import get_event_data
from eth_abi.codec import ABICodec
import pickle

from snet.sdk.mpe.payment_channel import PaymentChannel
from snet.contracts import get_contract_deployment_block


BLOCKS_PER_BATCH = 5000
CHANNELS_DIR = Path.home().joinpath(".snet", "cache", "mpe")


class PaymentChannelProvider(object):
    def __init__(self, w3, mpe_contract):
        self.web3 = w3

        self.mpe_contract = mpe_contract
        self.event_topics = [self.web3.keccak(
            text="ChannelOpen(uint256,uint256,address,address,address,bytes32,uint256,uint256)")]
        self.deployment_block = get_contract_deployment_block(self.web3, "MultiPartyEscrow")
        self.mpe_address = mpe_contract.contract.address
        print(self.mpe_address)
        self.channels_file = CHANNELS_DIR.joinpath(str(self.mpe_address), "channels.pickle")

    def update_cache(self):
        channels = []
        last_read_block = self.deployment_block - 1

        if not self.channels_file.exists():
            print(f"Channels cache is empty. Caching may take some time when first accessing channels.\nCaching in progress...")
            self.channels_file.parent.mkdir(parents=True, exist_ok=True)
            with open(self.channels_file, "wb") as f:
                empty_dict = {
                    "last_read_block": last_read_block,
                    "channels": channels
                }
                pickle.dump(empty_dict, f)
        else:
            with open(self.channels_file, "rb") as f:
                load_dict = pickle.load(f)
            last_read_block = load_dict["last_read_block"]
            channels = load_dict["channels"]

        current_block_number = self.web3.eth.block_number

        if last_read_block < current_block_number:
            new_channels = self._get_all_channels_from_blockchain_logs_to_dicts(last_read_block + 1, current_block_number)
            channels = channels + new_channels
            last_read_block = current_block_number

            with open(self.channels_file, "wb") as f:
                dict_to_save = {
                    "last_read_block": last_read_block,
                    "channels": channels
                }
                pickle.dump(dict_to_save, f)

    def _event_data_args_to_dict(self, event_data):
        return {
            "channel_id": event_data["channelId"],
            "sender": event_data["sender"],
            "signer": event_data["signer"],
            "recipient": event_data["recipient"],
            "group_id": event_data["groupId"],
        }

    def _get_all_channels_from_blockchain_logs_to_dicts(self, starting_block_number, to_block_number):
        codec: ABICodec = self.web3.codec

        logs = []
        from_block = starting_block_number
        while from_block <= to_block_number:
            to_block = min(from_block + BLOCKS_PER_BATCH, to_block_number)
            logs = logs + self.web3.eth.get_logs({"fromBlock": from_block,
                                                  "toBlock": to_block,
                                                  "address": self.mpe_address,
                                                  "topics": self.event_topics})
            from_block = to_block + 1

        event_abi = self.mpe_contract.contract.events.ChannelOpen._get_event_abi()

        event_data_list = [get_event_data(codec, event_abi, l)["args"] for l in logs]
        channels_opened = list(map(self._event_data_args_to_dict, event_data_list))

        return channels_opened

    def _get_channels_from_cache(self):
        self.update_cache()
        with open(self.channels_file, "rb") as f:
            load_dict = pickle.load(f)
        return load_dict["channels"]

    def get_past_open_channels(self, account, payment_address, group_id, payment_channel_state_service_client):

        dict_channels = self._get_channels_from_cache()

        channels_opened = list(filter(lambda channel: (channel["sender"] == account.address
                                                       or channel["signer"] == account.signer_address)
                                                      and channel["recipient"] == payment_address
                                                      and channel["group_id"] == group_id,
                                      dict_channels))

        return list(map(lambda channel: PaymentChannel(channel["channel_id"],
                                                       self.web3,
                                                       account,
                                                       payment_channel_state_service_client,
                                                       self.mpe_contract),
                        channels_opened))

    def open_channel(self, account, amount, expiration, payment_address, group_id, payment_channel_state_service_client):
        receipt = self.mpe_contract.open_channel(account, payment_address, group_id, amount, expiration)
        return self._get_newly_opened_channel(account, payment_address, group_id, receipt, payment_channel_state_service_client)

    def deposit_and_open_channel(self, account, amount, expiration, payment_address, group_id, payment_channel_state_service_client):
        receipt = self.mpe_contract.deposit_and_open_channel(account, payment_address, group_id, amount, expiration)
        return self._get_newly_opened_channel(account, payment_address, group_id, receipt, payment_channel_state_service_client)

    def _get_newly_opened_channel(self, account, payment_address, group_id, receipt, payment_channel_state_service_client):
        open_channels = self.get_past_open_channels(account, payment_address, group_id, payment_channel_state_service_client)
        if not open_channels:
            raise Exception(f"Error while opening channel, please check transaction {receipt.transactionHash.hex()} ")
        return open_channels[-1]




import web3
from snet.sdk.payment_strategies.payment_strategy import PaymentStrategy


class PaidCallPaymentStrategy(PaymentStrategy):
    def __init__(self, block_offset=240, call_allowance=1):
        self.block_offset = block_offset
        self.call_allowance = call_allowance

    def get_price(self, service_client):
        return service_client.get_price()

    def get_payment_metadata(self, service_client):
        channel = self.select_channel(service_client)
        amount = channel.state["last_signed_amount"] + int(self.get_price(service_client))
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256", "uint256", "uint256"],
            ["__MPE_claim_message", service_client.mpe_address, channel.channel_id,
             channel.state["nonce"],
             amount]
        )
        signature = service_client.generate_signature(message)

        metadata = [
            ("snet-payment-type", "escrow"),
            ("snet-payment-channel-id", str(channel.channel_id)),
            ("snet-payment-channel-nonce", str(channel.state["nonce"])),
            ("snet-payment-channel-amount", str(amount)),
            ("snet-payment-channel-signature-bin", signature)
        ]

        return metadata

    def select_channel(self, service_client):
        account = service_client.account
        service_client.load_open_channels()
        service_client.update_channel_states()
        payment_channels = service_client.payment_channels
        # picking the first pricing strategy as default for now
        service_call_price = self.get_price(service_client)
        mpe_balance = account.escrow_balance()
        default_expiration = service_client.default_channel_expiration()

        if len(payment_channels) < 1:
            if service_call_price > mpe_balance:
                payment_channel = service_client.deposit_and_open_channel(service_call_price,
                                                                          default_expiration + self.block_offset)
            else:
                payment_channel = service_client.open_channel(service_call_price,
                                                              default_expiration + self.block_offset)
            service_client.payment_channels = service_client.payment_channels + [payment_channel]
            service_client.update_channel_states()
        else:
            payment_channel = payment_channels[0]

        if self._has_sufficient_funds(payment_channel, service_call_price) and not self._is_valid(payment_channel,
                                                                                                  default_expiration):
            payment_channel.extend_expiration(default_expiration + self.block_offset)
        elif not self._has_sufficient_funds(payment_channel, service_call_price) and self._is_valid(payment_channel,
                                                                                                    default_expiration):
            payment_channel.add_funds(service_call_price * self.call_allowance)
        elif not self._has_sufficient_funds(payment_channel, service_call_price) and not self._is_valid(payment_channel,
                                                                                                        default_expiration):
            payment_channel.extend_and_add_funds(default_expiration + self.block_offset,
                                                 service_call_price * self.call_allowance)

        return payment_channel

    @staticmethod
    def _has_sufficient_funds(channel, amount):
        return channel.state["available_amount"] >= amount

    @staticmethod
    def _is_valid(channel, expiry):
        return channel.state["expiration"] >= expiry
